function plot(env, P::Project)
  plt_eq   = plot_equation(env, P)
  plt_rhs  = nothing # plot_rhs(env, P)
  plt_hrsc = plot_hrsc(env, P, P.hrsc)
  plt_tci  = plot_tci(env, P, P.tci)

  list_sub_plts = [ plt_eq, plt_rhs, plt_hrsc, plt_tci ]
  filter!(!isnothing, list_sub_plts)
  layout = (length(list_sub_plts), 1)
  plt_heights = [ 0.7, [ 0.3 / length(list_sub_plts[2:end]) for _ in 1:length(list_sub_plts[2:end]) ]...]
  sub_plt = Plots.plot(list_sub_plts..., layout=grid(layout..., heights=plt_heights))

  # combine plots
  @unpack cache = env
  @unpack t = get_global_variables(cache)
  plt = Plots.plot(sub_plt, layout=layout, size=(1400,1000), plot_title="t=$(t[1])",
                   left_margin=10Plots.mm)
  display(plt)
end


#######################################################################
#                              Equation                               #
#######################################################################


function plot_equation(env, P)

  @unpack cache, mesh    = env
  @unpack D, S, tau      = get_dynamic_variables(cache)
  @unpack rho, p, eps, v = get_static_variables(cache)
  @unpack t              = get_global_variables(cache)
  @unpack x              = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  L       = layout(mesh)
  _, K    = L

  D = reshape(D, L)
  S = reshape(S, L)
  tau = reshape(tau, L)
  rho = reshape(rho, L)
  p   = reshape(p, L)
  v   = reshape(v, L)
  eps = reshape(eps, L)
  x   = reshape(x, L)

  vars, ylabels = @stringify_names [ D, S, tau, rho, p, v, eps, nothing ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    isnothing(var) && continue
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(4,2))
 
  return plt
end


#######################################################################
#                                HRSC                                 #
#######################################################################


function plot_hrsc(env, P, hrsc::Nothing) end
function plot_hrsc(env, P, hrsc::AbstractReconstruction) end


function plot_hrsc(env, P, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache, mesh = env
  @unpack mu          = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "μ"
  L       = layout(mesh)
  _, K = L

  mu = ones(L) .* mu'
  x  = reshape(x, L)

  plt = Plots.plot(xlabel=xlabel, ylabel=ylabel, title="Artificial viscosity")
  for k in 1:K
    Plots.plot!(plt, x[:,k], mu[:,k], label=false)
  end

  return plt
end


function plot_hrsc(env, P, hrsc::HRSC.SmoothedArtificialViscosity)

  plt_mu = plot_hrsc(env, P, hrsc.av)

  @unpack cache, mesh = env
  @unpack smoothed_mu = get_static_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "μ"
  L       = layout(mesh)
  _, K    = L

  smoothed_mu = reshape(smoothed_mu, L)
  x           = reshape(x, L)

  plt_smoothed_mu = Plots.plot(xlabel=xlabel, ylabel=ylabel,
                               title="Smoothed Artificial Viscosity")
  for k in 1:K
    Plots.plot!(plt_smoothed_mu, x[:,k], smoothed_mu[:,k], label=false)
  end

  combined_plt = Plots.plot(plt_mu, plt_smoothed_mu, layout=(1,2))

  return combined_plt
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function plot_tci(env, P, tci::Nothing) end


function plot_tci(env, P, tci::TCI.AbstractTCI)

  @unpack cache, mesh = env
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylimits = (-0.5,1.5)
  ylabel  = "Entropy production"
  L       = layout(mesh)
  _, K    = L

  flag = ones(L) .* flag'
  x    = reshape(x, L)

  vars    = [ flag ]
  ylabels = [ "flag" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i],
                         xlimits=xlimits, ylimits=ylimits) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt
end


function plot_tci(env, P, tci::TCI.EntropyProduction)

  @unpack cache, mesh = env
  @unpack EP          = get_static_variables(cache)
  @unpack flag        = get_cell_variables(cache)
  @unpack x           = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  ylabel  = "Entropy production"
  L       = layout(mesh)
  _, K = L

  flag = ones(L) .* flag'
  EP   = reshape(EP, L)
  x    = reshape(x, L)

  vars    = [ flag, EP ]
  ylabels = [ "flag", "|∂t S + ∂x F_S|" ]
  titles  = [ "Troubled Cell Indicator", "Entropy Production" ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i], title=titles[i])
              for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(1,length(vars)))

  return plt
end


#######################################################################
#                                 RHS                                 #
#######################################################################


function plot_rhs(env, P)

  @unpack cache, mesh = env
  @unpack rhs_rho, rhs_q, rhs_E = get_rhs_variables(cache)
  @unpack t                     = get_global_variables(cache)
  @unpack x                     = mesh

  xlabel  = "x"
  xlimits = extrema(x)
  L       = layout(mesh)
  _, K    = L

  rhs_rho = reshape(rhs_rho, L)
  rhs_q   = reshape(rhs_q, L)
  rhs_E   = reshape(rhs_E, L)
  x   = reshape(x, L)

  vars    = [ rhs_rho, rhs_q, rhs_E, nothing ]
  ylabels = [ "rhs_rho", "rhs_q", "rhs_E", nothing ]
  plts    = [ Plots.plot(xlabel=xlabel, ylabel=ylabels[i]) for i in 1:length(vars) ]
  for (var, plt) in zip(vars, plts)
    isnothing(var) && continue
    for k in 1:K
      Plots.plot!(plt, x[:,k], var[:,k], label=false)
    end
  end

  plt = Plots.plot(plts..., layout=(2,2))

  return plt
end
