#######################################################################
#                              Timestep                               #
#######################################################################

timestep(env, P, hrsc) = timestep(env, P, hrsc, env.mesh)


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P
  @unpack max_v = get_static_variables(cache)

  # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)
  dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  dg1d.new_broadcast_volume!(new_speed, equation, mesh)
  vmax = dg1d.absolute_maximum(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P

  # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)
  dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  dg1d.new_broadcast_volume!(new_speed, equation, mesh)
  max_v = get_variable(cache, :max_v)
  vmax = dg1d.absolute_maximum(view(max_v,:))
  vmax_limit = 0.99
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                                 2d                                  #
#######################################################################


function timestep(env, P, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh2d)
  @unpack cache    = env
  @unpack equation = P
  @unpack max_v = get_static_variables(cache)

  TODO("Update cons2prim call here")
  broadcast_volume_2!(cons2prim, equation, mesh)
  broadcast_volume_2!(maxspeed, equation, mesh)
  i_vmax, vmax = dg1d.absolute_maximum_index(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    i_cell = find_cell_index(i_vmax, mesh)
    cell = mesh.tree.cells[i_cell]
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit in cell '$cell'"
    vmax = vmax_limit
  end

  @unpack CFL, element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = CFL * dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


rhs!(env, P, hrsc, bdryconds, ldg_bdryconds, av_bdryconds) =
  rhs!(env, P, hrsc, bdryconds, ldg_bdryconds, av_bdryconds, env.mesh)

function rhs!(env, P::Project, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds, mesh::Mesh1d)

  @unpack cache, mesh              = env
  @unpack equation, rsolver        = P
  @unpack D, S, tau                = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau    = get_rhs_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          max_v, rho, eps, v, p    = get_static_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_rho, bdry_v, bdry_eps,
          bdry_p, bdry_max_v       = get_bdry_variables(cache)

  # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)
  dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  dg1d.new_broadcast_volume!(new_speed, equation, mesh)
  dg1d.new_broadcast_volume!(flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D, bdry_D)
  dg1d.interpolate_face_data!(mesh, S, bdry_S)
  dg1d.interpolate_face_data!(mesh, tau, bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)

  dg1d.new_broadcast_faces!(llf, equation, mesh)
  dg1d.new_broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractReconstruction,
    bdryconds, ldg_bdryconds, av_bdryconds, ::Mesh1d)

  @unpack cache, mesh           = env
  @unpack equation, rsolver     = P
  @unpack D, S, tau             = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau = get_rhs_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          max_v, rho, eps, v, p = get_static_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_rho, bdry_v, bdry_eps,
          bdry_p, bdry_max_v    = get_bdry_variables(cache)
  @unpack flag                  = get_cell_variables(cache)


  isperiodic(x) = true
  HRSC.reconstruct!(D,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(S,    flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(tau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)
  dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  # dg1d.new_broadcast_volume!(new_speed, equation, mesh)
  dg1d.new_broadcast_volume!(flux, equation, mesh)
  # broadcast_volume!(cons2prim, equation, cache)
  # broadcast_volume!(flux, equation, cache)
  # broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  dg1d.interpolate_face_data!(mesh, D, bdry_D)
  dg1d.interpolate_face_data!(mesh, S, bdry_S)
  dg1d.interpolate_face_data!(mesh, tau, bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)
  dg1d.new_broadcast_faces!(llf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds, ::Mesh1d)

  @unpack cache, mesh                       = env
  @unpack equation, rsolver, ldg_rsolver,
          av_rsolver                        = P
  @unpack D, S, tau                         = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          max_v, rho, eps, v, p,
          ldg_D, ldg_S, ldg_tau,
          smoothed_mu, flag_consfixed       = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau             = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau,
          bdry_D, bdry_S, bdry_tau,
          bdry_rho, bdry_v, bdry_eps,
          bdry_p, bdry_max_v,
          bdry_ldg_D, bdry_ldg_S,
          bdry_ldg_tau,
          bdry_smoothed_mu                  = get_bdry_variables(cache)
  @unpack mu                                = get_cell_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  dg1d.new_broadcast_faces!(ldg_nflux, equation, mesh)
  dg1d.new_broadcast_bdry!(ldg_bdryllf, equation, mesh)
  # broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_D,   D,   nflx_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   S,   nflx_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, tau, nflx_tau, mesh)


  # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)

  ## applying bernsteining where viscosity appeared cures many artificial oscillations
  ## occuring from the AV, but then the AV method appears to be useless...
  # prms = parameters(:HRSC)
  # prms["method"] = "filter"
  # prms["filter_method"] = "bernstein"
  # bernstein = HRSC.make_HRSC(mesh, prms)
  # flag = zeros(size(mu))
  # # Npts, K = mesh.element.Npts, mesh.tree.dims[1]
  # # for i = 1:Npts*K
  # #   ii = div(i, K) + 1
  # #   f = flag[ii]
  # #   if mu[ii] > f
  # #     flag[ii] = 1.0
  # #   end
  # # end
  # for i = 1:length(mu)
  #   flag[i] = mu[i] > 0 ? 0.1 : 0.0
  # end
  # HRSC.reconstruct!(D,    flag, bernstein, isperiodic=true, m=1e-10, limit_with_boundaries=false)
  # HRSC.reconstruct!(S,    flag, bernstein, isperiodic=true, limit_with_boundaries=false)
  # HRSC.reconstruct!(tau,  flag, bernstein, isperiodic=true, m=1e-10, limit_with_boundaries=false)

  # dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  # dg1d.new_broadcast_volume!(new_speed, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D, bdry_D)
  dg1d.interpolate_face_data!(mesh, S, bdry_S)
  dg1d.interpolate_face_data!(mesh, tau, bdry_tau)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, rho, bdry_rho)
  dg1d.interpolate_face_data!(mesh, v, bdry_v)
  dg1d.interpolate_face_data!(mesh, eps, bdry_eps)
  dg1d.interpolate_face_data!(mesh, p, bdry_p)

  # # # dg1d.new_broadcast_volume!(conservative_fixing, equation, mesh)
  dg1d.new_broadcast_volume!(cons2prim_kastaun, equation, mesh)
  dg1d.new_broadcast_volume!(new_speed, equation, mesh)

  dg1d.new_broadcast_volume!(flux, equation, mesh)
  dg1d.new_broadcast_volume!(av_flux, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D, bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_S, bdry_ldg_S)
  dg1d.interpolate_face_data!(mesh, ldg_tau, bdry_ldg_tau)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  # numerical fluxes
  dg1d.new_broadcast_faces!(llf, equation, mesh)
  dg1d.new_broadcast_faces!(av_nflux, equation, mesh)
  dg1d.new_broadcast_bdry!(bdryllf, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, nflx_tau, mesh)

  return
end


#######################################################################
#                                 2d                                  #
#######################################################################

function dg1d.rhs!(env::Environment, P::Project2d, bdryconds)
  dg1d.rhs!(env, P, P.hrsc, env.mesh)
end

function dg1d.rhs!(env, P, ::Nothing, ::Mesh2d)

  @unpack cache, mesh                              = env
  @unpack equation                                 = P
  @unpack D, Sx, Sy, tau                           = get_dynamic_variables(cache)
  @unpack flx_D_x, flx_D_y, flx_Sx_x, flx_Sx_y,
          flx_Sy_x, flx_Sy_y, flx_tau_x, flx_tau_y = get_static_variables(cache)
  @unpack rhs_D, rhs_Sx, rhs_Sy, rhs_tau           = get_rhs_variables(cache)
  @unpack nflx_D, nflx_Sx, nflx_Sy, nflx_tau       = get_bdry_variables(cache)

  broadcast_volume_2!(cons2prim, equation, mesh)
  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_D, equation, mesh)
  broadcast_faces_2!(llf_Sx, equation, mesh)
  broadcast_faces_2!(llf_Sy, equation, mesh)
  broadcast_faces_2!(llf_tau, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D_x,    flx_D_y,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_Sx,  flx_Sx_x,   flx_Sx_y,  nflx_Sx,  mesh)
  compute_rhs_weak_form!(rhs_Sy,  flx_Sy_x,   flx_Sy_y,  nflx_Sy,  mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau_x,  flx_tau_y, nflx_tau, mesh)

  return
end


function dg1d.rhs!(env, P::Project2d, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh2d)

  @unpack equation                       = P
  @unpack D, Sx, Sy, tau                 = get_dynamic_variables(mesh.cache)
  @unpack rhs_D, rhs_Sx, rhs_Sy, rhs_tau = get_rhs_variables(mesh.cache)
  @unpack flx_D_x, flx_D_y,
          flx_Sx_x, flx_Sx_y,
          flx_Sy_x, flx_Sy_y,
          flx_tau_x, flx_tau_y,
          ldg_D_x, ldg_D_y,
          ldg_Sx_x, ldg_Sx_y,
          ldg_Sy_x, ldg_Sy_y,
          ldg_tau_x, ldg_tau_y               = get_static_variables(mesh.cache)
  @unpack nflx_D, nflx_Sx, nflx_Sy, nflx_tau,
          nflx_D_x, nflx_D_y,
          nflx_Sx_x, nflx_Sx_y,
          nflx_Sy_x, nflx_Sy_y,
          nflx_tau_x, nflx_tau_y             = get_bdry_variables(mesh.cache)
  @unpack t = get_global_variables(mesh.cache)

  ## TODO Maybe implement a macro with this syntax
  # @weakform(mesh) ∂t ldg_D_x + ∂x D = [ nflx_D_x ]
  # -> compute_rhs_weak_form(ldg_D_x, D, 0, nflx_D_x, mesh)
  # @weakform(mesh) ∂t ldg_D_y + ∂y D = [ nflx_D_y ]
  # @weakform(mesh) ∂t ldg_Sx_x  + ∂x Sx  = [ nflx_Sx_x ]
  # @weakform(mesh) ∂t ldg_Sx_y  + ∂y Sx  = [ nflx_Sx_y ]
  # @weakform(mesh) ∂t ldg_Sy_x  + ∂x Sy  = [ nflx_Sy_x ]
  # @weakform(mesh) ∂t ldg_Sy_y  + ∂y Sy  = [ nflx_Sy_y ]
  # @weakform(mesh) ∂t ldg_tau_x   + ∂x tau   = [ nflx_tau_x ]
  # @weakform(mesh) ∂t ldg_tau_y   + ∂y tau   = [ nflx_tau_y ]

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_faces_2!(ldg_nflux, equation, mesh)
  dg1d.broadcast_bdry_2!(bdry_ldg_nflux, equation, P.bdrycond, mesh)

  # broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_D_x,   D,   0,   nflx_D_x,   mesh)
  compute_rhs_weak_form!(ldg_D_y,   0,   D,   nflx_D_y,   mesh)
  compute_rhs_weak_form!(ldg_Sx_x,  Sx,  0,   nflx_Sx_x,  mesh)
  compute_rhs_weak_form!(ldg_Sx_y,  0,   Sx,  nflx_Sx_y,  mesh)
  compute_rhs_weak_form!(ldg_Sy_x,  Sy,  0,   nflx_Sy_x,  mesh)
  compute_rhs_weak_form!(ldg_Sy_y,  0,   Sy,  nflx_Sy_y,  mesh)
  compute_rhs_weak_form!(ldg_tau_x, tau, 0,   nflx_tau_x, mesh)
  compute_rhs_weak_form!(ldg_tau_y, 0,   tau, nflx_tau_y, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume_2!(cons2prim, equation, mesh)
  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf, equation, mesh)
  dg1d.broadcast_bdry_2!(bdry_llf, equation, P.bdrycond, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  broadcast_volume_2!(av_flux_2d, equation, mesh)
  broadcast_faces_2!(av_nflux_2d, equation, mesh)
  dg1d.broadcast_bdry_2!(bdry_av_nflux_2d, equation, P.bdrycond, mesh)
  # broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
  # broadcast_faces_2!(mean_flux, equation, mesh, P.bdryconds)

  compute_rhs_weak_form!(rhs_D,   flx_D_x,   flx_D_y,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_Sx,  flx_Sx_x,  flx_Sx_y,  nflx_Sx,  mesh)
  compute_rhs_weak_form!(rhs_Sy,  flx_Sy_x,  flx_Sy_y,  nflx_Sy,  mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau_x, flx_tau_y, nflx_tau, mesh)

  return
end
